from sanic.views import HTTPMethodView
from sanic.response import json, file
from naas.types import t_static
import uuid


class LogsController(HTTPMethodView):
    __logger = None

    def __init__(self, logger, *args, **kwargs):
        super(LogsController, self).__init__(*args, **kwargs)
        self.__logger = logger

    async def get(self, request):
        # return json({'data': [], 'totalRecords': 0})
        as_file = request.args.get("as_file", False)
        if as_file:
            return await file(
                self.__logger.get_file_path(), attachment_filename="logs.csv"
            )
        else:
            uid = str(uuid.uuid4())
            limit = int(request.args.get("limit", 0))
            skip = int(request.args.get("skip", 0))
            search = str(request.args.get("search", ""))
            filters = list(request.args.get("filters", []))
            logs = self.__logger.list(uid, skip, limit, search, filters)
            self.__logger.info(
                {
                    "id": uid,
                    "type": t_static,
                    "status": "send",
                    "filepath": "logs",
                    "skip": skip,
                    "limit": limit,
                    "search": search,
                    "filters": filters,
                }
            )
            return json(logs)



# from gevent.lock import BoundedSemaphore
from asyncio import Semaphore
from naas.types import t_delete, t_add, t_skip, t_update, t_error, t_start
import pandas as pd
import datetime
import errno
import json
import os
import uuid


class Jobs:
    __storage_sem = None
    __path_user_files = None
    __logger = None
    __naas_folder = ".naas"
    __json_name = "jobs.json"

    def __init__(self, logger, clean=False, init_data=[]):
        self.__path_user_files = os.environ.get("JUPYTER_SERVER_ROOT", "/home/ftp")
        self.__path_naas_files = os.path.join(
            self.__path_user_files, self.__naas_folder
        )
        self.__json_secrets_path = os.path.join(
            self.__path_naas_files, self.__json_name
        )
        # self.__storage_sem = BoundedSemaphore(1)
        self.__storage_sem = Semaphore(1)
        self.__logger = logger
        if not os.path.exists(self.__path_naas_files):
            try:
                print("Init Naas folder Jobs")
                os.makedirs(self.__path_naas_files)
            except OSError as exc:  # Guard against race condition
                print("__path_naas_files", self.__path_naas_files)
                if exc.errno != errno.EEXIST:
                    raise
            except Exception as e:
                print("Exception", e)
        if not os.path.exists(self.__json_secrets_path) or clean:
            uid = str(uuid.uuid4())
            try:
                print("Init Job Storage", self.__json_secrets_path)
                self.__save(uid, init_data)
            except Exception as e:
                print("Exception", e)
                self.__logger.error(
                    {
                        "id": uid,
                        "type": "init_job_storage",
                        "status": "error",
                        "error": str(e),
                    }
                )

    def __save(self, uid, data):
        try:
            with open(self.__json_secrets_path, "w+") as f:
                f.write(
                    json.dumps(data, sort_keys=True, indent=4).replace("NaN", "null")
                )
                f.close()
        except Exception as err:
            print("__save Exception", err)
            self.__logger.error(
                {
                    "id": str(uid),
                    "type": "set_job_storage",
                    "status": "exception",
                    "filepath": self.__json_secrets_path,
                    "error": str(err),
                }
            )

    def find_by_value(self, uid, value, target_type):
        data = self.list(uid)
        if len(data) > 0:
            job_list = pd.DataFrame(data)
            cur_jobs = job_list[
                (job_list.type == target_type) & (job_list.value == value)
            ]
            cur_job = cur_jobs.to_dict("records")
            if len(cur_job) == 1:
                return cur_job[0]
        return None

    def find_by_path(self, uid, filepath, target_type):
        data = self.list(uid)
        if len(data) > 0:
            job_list = pd.DataFrame(data)
            cur_jobs = job_list[
                (job_list.type == target_type) & (job_list.path == filepath)
            ]
            cur_job = cur_jobs.to_dict("records")
            if len(cur_job) == 1:
                return cur_job[0]
        return None

    def is_running(self, uid, notebook_filepath, target_type):
        cur_job = self.find_by_path(uid, notebook_filepath, target_type)
        if cur_job:
            status = cur_job.get("status", None)
            if status and status == t_start:
                return True
        return False

    def list(self, uid):
        data = []
        try:
            with open(self.__json_secrets_path, "r") as f:
                data = json.load(f)
                f.close()
        except Exception as err:
            print("cannot open")
            self.__logger.error(
                {
                    "id": uid,
                    "type": "get_job_storage",
                    "status": "exception",
                    "filepath": self.__json_secrets_path,
                    "error": str(err),
                }
            )
            data = []
        return data

    async def update(self, uid, path, target_type, value, params, status, runTime=0):
        await self.__storage_sem.acquire()
        data = None
        res = t_error
        try:
            if len(self.list(uid)) != 0:
                df = pd.DataFrame(self.list(uid))
            else:
                df = pd.DataFrame(
                    columns=[
                        "id",
                        "type",
                        "value",
                        "path",
                        "status",
                        "params",
                        "lastUpdate",
                        "lastRun",
                        "totalRun",
                    ]
                )
            res = status
            cur_elem = df[(df.type == target_type) & (df.path == path)]
            now = datetime.datetime.now()
            dt_string = now.strftime("%Y-%m-%d %H:%M:%S")
            if len(cur_elem) == 1:
                if status == t_delete:
                    self.__logger.info(
                        {
                            "id": uid,
                            "type": target_type,
                            "value": value,
                            "status": t_delete,
                            "path": path,
                            "params": params,
                        }
                    )
                    df = df.drop(cur_elem.index)
                else:
                    self.__logger.info(
                        {
                            "id": uid,
                            "type": target_type,
                            "value": value,
                            "status": t_update,
                            "path": path,
                            "params": params,
                        }
                    )
                    index = cur_elem.index[0]
                    df.at[index, "id"] = uid
                    df.at[index, "status"] = status
                    df.at[index, "value"] = value
                    df.at[index, "params"] = params
                    df.at[index, "lastUpdate"] = dt_string
                    if runTime > 0 and status != t_add:
                        df.at[index, "lastRun"] = runTime
                        df.at[index, "totalRun"] = runTime + (
                            df.at[index, "totalRun"] if df.at[index, "totalRun"] else 0
                        )
                    elif status == t_add:
                        df.at[index, "lastRun"] = 0
                        df.at[index, "totalRun"] = 0
                    res = t_update
            elif status == t_add and len(cur_elem) == 0:
                self.__logger.info(
                    {
                        "id": uid,
                        "type": target_type,
                        "value": value,
                        "status": t_update,
                        "path": path,
                        "params": params,
                    }
                )
                new_row = [
                    {
                        "id": uid,
                        "type": target_type,
                        "value": value,
                        "status": t_add,
                        "path": path,
                        "params": params,
                        "lastRun": runTime,
                        "totalRun": runTime,
                        "lastUpdate": dt_string,
                    }
                ]
                df_new = pd.DataFrame(new_row)
                df = pd.concat([df, df_new], axis=0)
            else:
                res = t_skip
            data = df.to_dict("records")
            if res != t_skip:
                self.__save(uid, data)
        except Exception as e:
            print("cannot update", e)
            self.__logger.error(
                {
                    "id": uid,
                    "type": target_type,
                    "value": value,
                    "status": t_error,
                    "path": path,
                    "params": params,
                    "error": str(e),
                }
            )
        self.__storage_sem.release()
        return {"status": res, "data": data}



from sentry_sdk.integrations.sanic import SanicIntegration
from naas.runner.controllers.scheduler import SchedulerController
from naas.runner.controllers.static import StaticController
from naas.runner.controllers.notebooks import NbController
from naas.runner.controllers.jobs import JobsController
from naas.runner.controllers.logs import LogsController
from naas.runner.controllers.env import EnvController
from naas.runner.notifications import Notifications
from naas.runner.scheduler import Scheduler
from naas.runner.notebooks import Notebooks
from sanic_openapi import swagger_blueprint
from naas.runner.logger import Logger
from .proxy import escape_kubernet
from naas.runner.jobs import Jobs
from daemonize import Daemonize
from naas.types import t_main
from sanic import Sanic
import sentry_sdk
import asyncio
import getpass
import uuid
import os
import json
import sys


class Runner:
    # Declare semaphore variable
    __name = "naas_runner"
    __naas_folder = ".naas"
    __tasks_sem = None
    # Declare path variable
    __path_lib_files = os.path.dirname(os.path.abspath(__file__))
    __path_user_files = None
    __port = 5000
    __single_user_api_path = None
    __html_files = "html"
    __assets_files = "assets"
    __manager_index = "manager.html"
    __log_filename = "logs.csv"
    __tasks_files = "jobs.json"
    __info_file = "info.json"
    __app = None
    __nb = None
    __http_server = None
    __notif = None
    __jobs = None
    __scheduler = None
    __daemon = None
    __sentry = None
    __logger = None
    __testing = False
    tz = None
    user = None
    __shell_user = None
    public_url = None
    proxy_url = None

    def __init__(self):
        self.__path_user_files = os.environ.get("JUPYTER_SERVER_ROOT", "/home/ftp")
        self.__port = int(os.environ.get("NAAS_RUNNER_PORT", 5000))
        self.__user = os.environ.get("JUPYTERHUB_USER", "joyvan@naas.com")
        self.__shell_user = os.environ.get("USER", None)
        self.__public_url = os.environ.get("JUPYTERHUB_URL", "")
        self.__proxy_url = os.environ.get("PUBLIC_PROXY_API", "http://localhost:3002")
        self.__notifications_url = os.environ.get(
            "NOTIFICATIONS_API", "http://localhost:3003"
        )
        self.__tz = os.environ.get("TZ", "Europe/Paris")
        self.__path_naas_files = os.path.join(
            self.__path_user_files, self.__naas_folder
        )
        self.__path_html_files = os.path.join(self.__path_lib_files, self.__html_files)
        self.__path_manager_index = os.path.join(
            self.__path_html_files, self.__manager_index
        )
        self.__path_pidfile = os.path.join(self.__path_naas_files, f"{self.__name}.pid")

    def __main(self, debug=False):
        uid = str(uuid.uuid4())
        self.__logger.info({"id": uid, "type": t_main, "status": "start API"})
        self.__app.run(host="0.0.0.0", port=self.__port, debug=debug, access_log=debug)

    def __version(self):
        version = None
        try:
            with open(
                os.path.join(self.__path_lib_files, self.__info_file), "r"
            ) as json_file:
                version = json.load(json_file)
        except IOError:
            version = {"error": "cannot get info.json"}
        return version

    async def initialize_before_start(self, app, loop):
        if self.__nb is None:
            self.__nb = Notebooks(self.__logger, loop, self.__notif)
            self.__app.add_route(
                NbController.as_view(self.__logger, self.__jobs, self.__nb),
                "/notebooks/<token>",
            )
            self.__scheduler = Scheduler(self.__logger, self.__jobs, self.__nb, loop)
            self.__app.add_route(
                SchedulerController.as_view(self.__logger, self.__scheduler),
                "/scheduler/<mode>",
            )
            self.__scheduler.start()

    def kill(self):
        if os.path.exists(self.__path_pidfile):
            try:
                with open(self.__path_pidfile, "r") as f:
                    pid = f.read()
                    print("kill")
                    os.system(f"kill {pid}")
                    print(f"Deamon killed pid {pid}")
                    f.close()
            except Exception:
                print("No Deamon running")

    def init_app(self):
        self.__logger = Logger()
        self.__notif = Notifications(self.__logger)
        self.__jobs = Jobs(self.__logger)
        self.__app = Sanic(__name__)
        # TODO remove this fix when papermill support uvloop of Sanic support option to don't use uvloop
        asyncio.set_event_loop_policy(None)
        self.__app.register_listener(
            self.initialize_before_start, "before_server_start"
        )
        self.__app.add_route(
            EnvController.as_view(
                self.__logger,
                self.__user,
                self.__public_url,
                self.__proxy_url,
                self.__notifications_url,
                self.__tz,
            ),
            "/env",
        )
        self.__app.add_route(LogsController.as_view(self.__logger), "/logs")
        self.__app.add_route(
            JobsController.as_view(self.__logger, self.__jobs), "/jobs"
        )
        self.__app.add_route(
            StaticController.as_view(self.__logger, self.__jobs, self.__path_lib_files),
            "/static/<token>",
        )
        self.__app.static("/", self.__path_manager_index, name="manager.html")
        self.__app.blueprint(swagger_blueprint)
        uid = str(uuid.uuid4())
        self.__logger.info({"id": uid, "type": t_main, "status": "init API"})
        return self.__app

    def start(self, deamon=True, port=None, debug=True):
        user = getpass.getuser()
        if user != self.__shell_user:
            raise Exception(f"{user} not autorized, use {self.__shell_user} instead")
        self.kill()
        if port:
            self.__port = port
        self.init_app()
        if deamon:
            if os.environ.get("NAAS_SENTRY_DSN"):
                self.__sentry = sentry_sdk.init(
                    dsn=os.environ.get("NAAS_SENTRY_DSN"),
                    traces_sample_rate=1.0,
                    environment=escape_kubernet(self.__user),
                    integrations=[SanicIntegration()],
                )
            self.__daemon = Daemonize(
                app=self.__name, pid=self.__path_pidfile, action=self.__main
            )
            print("Start Runner Deamon Mode")
            self.__daemon.start()
        else:
            print("Start Runner front Mode")
            try:
                self.__main(debug)
            except KeyboardInterrupt:
                print("Shutdown server")
                sys.exit()



# Copyright (c) Naas Team.
# Distributed under the terms of the Modified BSD License.
from .refresh import Refresh
from .api import Api
from .static import Static
from .dependency import Dependency
from IPython.core.display import display, HTML
from .secret import Secret
from .runner import Runner
from .runner.proxy import encode_proxy_url
import requests
import os


__location__ = os.getcwd()
refresh = Refresh
secret = Secret
runner = Runner
api = Api
static = Static
dependency = Dependency


def manager():
    public_url = f"{encode_proxy_url()}"
    print("You can check your current tasks list here :")
    display(HTML(f'<a href="{public_url}"">Manager</a>'))


def refresh_status():
    req = requests.get(url="localhost:5000/scheduler")
    req.raise_for_status()
    jsn = req.json()
    print(jsn)
    return jsn


def refresh_pause():
    req = requests.get(url="localhost:5000/scheduler/pause")
    req.raise_for_status()
    jsn = req.json()
    print(jsn)
    return jsn


def refresh_resume():
    req = requests.get(url="localhost:5000/scheduler/resume")
    req.raise_for_status()
    jsn = req.json()
    print(jsn)
    return jsn


def welcome():
    print("[Nass from JupyterHub] => Welcome to your ETL, read the documentation")
    print("Do naas.help() to get more info on what you can do.\n")


def help():
    print("naas.version() => Get the current version\n")
    print("naas.manager() => Get the url of the manager\n")
    print("naas.refresh => Get the notebook refresh driver\n")
    print("naas.api => Get the notebook api driver\n")
    print("naas.static => Get the file sharing driver\n")
    print("naas.dependency => Get the file dependency driver\n")



t_notebook = "api"
t_static = "static"
t_task = "tasks"
t_dependency = "dependency"
t_scheduler = "scheduler"

t_add = "installed"
t_delete = "delete"
t_update = "edited"
t_start = "started"
t_skip = "skiped"
t_error = "error"
t_health = "healthy"
t_main = "main"



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="naas",
    version="0.1.0",
    scripts=["scripts/naas"],
    author="Martin Donadieu",
    author_email="martindonadieu@gmail.com",
    description="scheduler system for notebooks",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/cashstory/naas",
    packages=find_packages(exclude=["tests"]),
    package_data={
        "naas": ["runner/html/*.html", "runner/assets/*.png", "runner/assets/*.svg"],
    },
    tests_require=[
        "pytest==5.2",
        "pytest-sanic==1.6.1",
        "requests-mock==1.8.0",
        "twine==3.2.0",
        "flake8==3.1.1",
        "black==20.8b1",
        "commitizen==2.4.0",
    ],
    install_requires=[
        "versioneer",
        "papermill>=2,<3",
        "pretty-cron>=1,<2",
        "APScheduler>=3,<4",
        "pycron>=3,<4",
        "pandas>=1,<2",
        "daemonize>=2,<3",
        "escapism>=1,<2",
        "notebook>=6,<7",
        "ipython>=7,<8",
        "ipykernel>=5,<6",
        "requests>=2,<3",
        "sentry-sdk>=0,<1",
        "sanic>=20,<21",
        "sanic-openapi>=0,<1",
        "argparse>=1,<2",
        "nbconvert>=6,<7",
        "nbclient>=0,<1",
        "beautifulsoup4>=4,<5",
    ],
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



from naas.types import t_notebook, t_health, t_error, t_start
from sanic.views import HTTPMethodView
from sanic.response import json
import uuid


class NbController(HTTPMethodView):
    __logger = None

    def __init__(self, logger, jobs, nb, *args, **kwargs):
        super(NbController, self).__init__(*args, **kwargs)
        self.__logger = logger
        self.__jobs = jobs
        self.__nb = nb

    async def get(self, request, token):
        uid = str(uuid.uuid4())
        task = self.__jobs.find_by_value(uid, token, t_notebook)
        if task:
            value = task.get("value", None)
            file_filepath = task.get("path")
            task["params"] = {**(task.get("params", dict())), **(request.args)}
            self.__logger.info(
                {
                    "id": uid,
                    "type": t_notebook,
                    "status": t_start,
                    "filepath": file_filepath,
                    "token": token,
                }
            )
            res = await self.__nb.exec(uid, task)
            if res.get("error"):
                self.__logger.error(
                    {
                        "main_id": uid,
                        "id": uid,
                        "type": t_notebook,
                        "status": t_error,
                        "filepath": file_filepath,
                        "duration": res.get("duration"),
                        "error": res.get("error"),
                    }
                )
                await self.__jobs.update(
                    uid,
                    file_filepath,
                    t_notebook,
                    value,
                    task.get("params"),
                    t_error,
                    res.get("duration"),
                )
                return json({"id": uid, "error": res.get("error")})
            self.__logger.info(
                {
                    "main_id": uid,
                    "id": uid,
                    "type": t_notebook,
                    "status": t_health,
                    "filepath": file_filepath,
                    "duration": res.get("duration"),
                }
            )
            await self.__jobs.update(
                uid,
                file_filepath,
                t_notebook,
                value,
                task.get("params"),
                t_health,
                res.get("duration"),
            )
            return self.__nb.response(uid, res, res.get("duration"), task.get("params"))
        self.__logger.error(
            {
                "id": uid,
                "type": t_notebook,
                "status": t_error,
                "token": token,
                "error": "Cannot find your token",
            }
        )
        return json({"id": uid, "error": "Cannot find your token"}, status=404)



from naas.types import t_scheduler, t_start, t_main, t_health, t_error
from apscheduler.schedulers.asyncio import AsyncIOScheduler
import apscheduler.schedulers.base
import traceback
import datetime
import asyncio
import pycron
import time
import uuid


class Scheduler:
    __scheduler = None
    __notebooks = None
    __logger = None
    __jobs = None

    def __init__(self, logger, jobs, notebooks, loop):
        self.__logger = logger
        self.__notebooks = notebooks
        self.__jobs = jobs
        self.__scheduler = AsyncIOScheduler({"event_loop": loop})
        self.state = self.__scheduler.state

    def status(self):
        if self.__scheduler.state == apscheduler.schedulers.base.STATE_RUNNING:
            return "running"
        elif self.__scheduler.state == apscheduler.schedulers.base.STATE_PAUSED:
            return "paused"

    def start(self):
        if self.__scheduler.state != apscheduler.schedulers.base.STATE_RUNNING:
            self.__scheduler.add_job(
                func=self.__scheduler_function,
                trigger="interval",
                seconds=60,
                max_instances=10,
            )
            self.__scheduler.start()
            uid = str(uuid.uuid4())
            self.__logger.info({"id": uid, "type": t_main, "status": "start SCHEDULER"})

    def __scheduler_greenlet(self, main_uid, current_time, task):
        value = task.get("value", None)
        current_type = task.get("type", None)
        file_filepath = task.get("path")
        params = task.get("params", dict())
        uid = str(uuid.uuid4())
        running = self.__jobs.is_running(uid, file_filepath, current_type)
        if (
            current_type == t_scheduler
            and value is not None
            and pycron.is_now(value, current_time)
            and not running
        ):
            self.__logger.info(
                {
                    "main_id": str(main_uid),
                    "id": uid,
                    "type": t_scheduler,
                    "status": t_start,
                    "filepath": file_filepath,
                }
            )
            self.__jobs.update(uid, file_filepath, t_scheduler, value, params, t_start)
            res = self.__exec_job(uid, task)
            if res.get("error"):
                self.__logger.error(
                    {
                        "main_id": str(main_uid),
                        "id": uid,
                        "type": t_scheduler,
                        "status": t_error,
                        "filepath": file_filepath,
                        "duration": res.get("duration"),
                        "error": str(res.get("error")),
                    }
                )
                self.__jobs.update(
                    uid,
                    file_filepath,
                    t_scheduler,
                    value,
                    params,
                    t_error,
                    res.get("duration"),
                )
                return
            self.__logger.info(
                {
                    "main_id": str(main_uid),
                    "id": uid,
                    "type": t_scheduler,
                    "status": t_health,
                    "filepath": file_filepath,
                    "duration": res.get("duration"),
                }
            )
            self.__jobs.update(
                uid,
                file_filepath,
                t_scheduler,
                value,
                params,
                t_health,
                res.get("duration"),
            )

    async def __scheduler_function(self):
        # Create unique for scheduling step (one step every minute)
        main_uid = str(uuid.uuid4())
        try:
            current_time = datetime.datetime.now()
            all_start_time = time.time()
            greelets = []
            # Write self.__scheduler init info in self.__logger.write
            self.__logger.info({"id": main_uid, "type": t_scheduler, "status": t_start})
            for job in self.__jobs.list(main_uid):
                g = asyncio.ensure_future(
                    self.__scheduler_greenlet(main_uid, current_time, job)
                )
                # g = gevent.spawn(self.__scheduler_greenlet,
                #                 main_uid, current_time, job)
                greelets.append(g)
            await asyncio.gather(*greelets)
            # gevent.joinall(greelets)
            durationTotal = time.time() - all_start_time
            self.__logger.info(
                {
                    "id": main_uid,
                    "type": t_scheduler,
                    "status": t_health,
                    "duration": durationTotal,
                }
            )
        except Exception as e:
            tb = traceback.format_exc()
            durationTotal = time.time() - all_start_time
            self.__logger.error(
                {
                    "id": main_uid,
                    "type": t_scheduler,
                    "status": t_error,
                    "duration": durationTotal,
                    "error": str(e),
                    "trace": tb,
                }
            )
